import urllib.parse
from json import JSONEncoder
from typing import Dict, Union

from nicovideo_api_client.api.v2.json_filter import JsonFilterOperator, JsonFilterTerm
from nicovideo_api_client.api.v2.limit import SnapshotSearchAPIV2Limit
from nicovideo_api_client.api.v2.simple_filter import SnapshotSearchAPIV2SimpleFilter


class SnapshotSearchAPIV2Filter:
    def __init__(self, query: Dict[str, str]):
        self._query: Dict[str, str] = query

    def simple_filter(self) -> SnapshotSearchAPIV2SimpleFilter:
        """
        絞り込みにシンプルな `filters` を利用することを宣言する。

        :return: フィルタリング要素指定オブジェクト
        """
        return SnapshotSearchAPIV2SimpleFilter(self._query)

    def json_filter(
            self,
            op: Union[JsonFilterOperator, JsonFilterTerm]
    ) -> SnapshotSearchAPIV2Limit:
        """
        `JsonFilterOperator` を使って表現した条件から、 `jsonFilter` を利用して絞り込む。

        :param op: 絞り込み条件
        :return: リクエスト上限設定オブジェクト
        """
        self._query["jsonFilter"] = urllib.parse.quote_plus(JSONEncoder().encode(op.json))
        return SnapshotSearchAPIV2Limit(self._query)



from datetime import datetime
from typing import Dict, List, Optional, Union


TypeOp = Dict[str, Union[str, bool, int, List['TypeOp'], 'TypeOp']]


class JsonFilterOperator:
    """
    `jsonFilter` による絞り込みの条件を表現するクラス。

    基本的にはこのオブジェクト1つで検索条件を表現するが、入子的に複数の検索条件が設定されることがある。
    """
    def __init__(self, json_: TypeOp):
        self.json: TypeOp = json_

    @staticmethod
    def unit(term: 'JsonFilterTerm') -> 'JsonFilterOperator':
        """
        単一の絞り込み要素のみで絞り込むことを明示的に宣言する。

        基本的に要素のみでの指定ができるため、利用する必要がない。

        Note: deprecated

        :param term: 絞り込み要素
        :return: 絞り込み条件オブジェクト
        """
        return JsonFilterOperator(term.json)

    @staticmethod
    def not_(
            term: 'JsonFilterOperator'
    ) -> 'JsonFilterOperator':
        """
        与えられた絞り込み要素の論理否定を表現する絞り込み条件。

        :param term: 否定したい絞り込み要素
        :return: 絞り込み条件オブジェクト
        """
        json_: TypeOp = {
            "type": "not",
            "filter": term.json
        }
        return JsonFilterOperator(json_)

    @staticmethod
    def _binary(
            op_name: str,
            left: 'JsonFilterOperator',
            right: 'JsonFilterOperator'
    ) -> 'JsonFilterOperator':
        json_: TypeOp = {
            "type": op_name,
            "filters": []
        }
        json_["filters"].append(left.json)
        json_["filters"].append(right.json)
        return JsonFilterOperator(json_)

    @staticmethod
    def or_(
            left: 'JsonFilterOperator',
            right: 'JsonFilterOperator'
    ) -> 'JsonFilterOperator':
        """
        2つの絞り込み要素のどちらかが正しいことを表す絞り込み条件。

        :param left: 絞り込み要素 1
        :param right: 絞り込み要素 2
        :return: 絞り込み条件オブジェクト
        """
        return JsonFilterOperator._binary("or", left, right)

    @staticmethod
    def and_(
            left: 'JsonFilterOperator',
            right: 'JsonFilterOperator'
    ) -> 'JsonFilterOperator':
        """
        2つの絞り込み要素の両方が正しいことを表す絞り込み条件。

        :param left: 絞り込み要素 1
        :param right: 絞り込み要素 2
        :return: 絞り込み条件オブジェクト
        """
        return JsonFilterOperator._binary("and", left, right)


class JsonFilterTerm(JsonFilterOperator):
    """
    `FieldType` と値の関係が正しいかを示す検索要素。

    `equal` と `range` の2種類がある。

    TODO: equal 絞り込みを実装する。
    """
    def __init__(self):
        super().__init__({})

    @staticmethod
    def set_range_view(
            from_: Optional[int] = None,
            to_: Optional[int] = None,
            include_lower: bool = True,
            include_upper: bool = True,
    ) -> 'JsonFilterTerm':
        """
        再生回数が与えられた範囲内かを調べる検索要素。

        :param from_: 再生回数の下限
        :param to_: 再生回数の上限
        :param include_lower: from_ を含むかどうか
        :param include_upper: to_ を含むかどうか
        :return: 絞り込み要素オブジェクト
        """
        term: JsonFilterTerm = JsonFilterTerm()
        if from_ is None and to_ is None:
            raise Exception("上限も下限も指定されていません")
        json_: Dict[str, Union[str, bool, int]] = {
            "type": "range",
            "field": "viewCounter"
        }
        if from_ is not None:
            json_["from"] = from_
        if to_ is not None:
            json_["to"] = to_
        if include_lower:
            json_["include_lower"] = include_lower
        if include_upper:
            json_["include_upper"] = include_upper
        term.json = json_
        return term

    @staticmethod
    def set_range_time(
            from_: Optional[datetime] = None,
            to_: Optional[datetime] = None,
            include_lower: bool = True,
            include_upper: bool = True,
    ) -> 'JsonFilterTerm':
        """
        投稿時刻が与えられた範囲内かを調べる検索要素。

        :param from_: 投稿時刻の最古
        :param to_: 投稿時刻の最新
        :param include_lower: from_ を含むかどうか
        :param include_upper: to_ を含むかどうか
        :return: 絞り込み要素オブジェクト
        """
        term: JsonFilterTerm = JsonFilterTerm()
        if from_ is None and to_ is None:
            raise Exception("開始時刻も終了時刻も指定されていません")
        json_: Dict[str, Union[str, bool]] = {
            "type": "range",
            "field": "startTime"
        }
        if from_ is not None:
            json_["from"] = from_.strftime('%Y-%m-%dT%H:%M:%S+09:00')
        if to_ is not None:
            json_["to"] = to_.strftime('%Y-%m-%dT%H:%M:%S+09:00')
        if include_lower:
            json_["include_lower"] = include_lower
        if include_upper:
            json_["include_upper"] = include_upper
        term.json = json_
        return term



from typing import Dict

from nicovideo_api_client.api.v2.filter import SnapshotSearchAPIV2Filter
from nicovideo_api_client.constants import FieldType, sort_types


class SnapshotSearchAPIV2Sort:
    def __init__(self, query: Dict[str, str]):
        self._query: Dict[str, str] = query

    def sort(self, sort_type: FieldType, reverse: bool = False) -> SnapshotSearchAPIV2Filter:
        """
        ソートタイプを指定する。

        :param sort_type: ソートに利用するフィールドタイプ
        :param reverse: True: 昇順, False (デフォルト): 降順
        :return: フィルター指定オブジェクト
        """
        if sort_type not in sort_types:
            raise Exception("不正なソートタイプ")
        self._query['_sort'] = ('+' if reverse else '-') + sort_type.value
        return SnapshotSearchAPIV2Filter(self._query)



from typing import Dict

from nicovideo_api_client.api.v2.field import SnapshotSearchAPIV2Fields
from nicovideo_api_client.constants import FieldType


class SnapshotSearchAPIV2Targets:
    def __init__(self, *targets: FieldType):
        list_targets = list(targets)
        if len(list_targets) == 0:
            raise Exception("targets が設定されていません")
        self._query: Dict[str, str] = {'targets': ','.join(map(lambda x: x.value, list_targets))}

    def query(self, keyword: str) -> SnapshotSearchAPIV2Fields:
        """
        検索クエリ(キーワード)を指定する。

        `クエリ文字列仕様 <https://site.nicovideo.jp/search-api-docs/snapshot#%EF%BC%8A1-
        %E3%82%AF%E3%82%A8%E3%83%AA%E6%96%87%E5%AD%97%E5%88%97%E4%BB%95%E6%A7%98>`_
        に沿った値を入力することで AND, OR 検索など複数のキーワードを含めた検索を行うことができる。

        TODO: AND・OR検索を実装する

        :param keyword: 検索するキーワード
        :return: レスポンスフィールドのタイプ指定オブジェクト
        """
        self._query['q'] = keyword
        return SnapshotSearchAPIV2Fields(self._query)
